/*  lbm converter released under gpl v3 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned char uchar;
typedef unsigned int uint;
typedef struct { uchar r,g,b; } rgb_t;

void error(char *s) {
  puts(s);
  exit(1);
}

/*  read byte, 2byte, 4byte with big endian */
uint read4(uchar *p) { return ((uint)p[0]<<24)+((uint)p[1]<<16)+((uint)p[2]<<8)+(uint)p[3]; }
int read2(uchar *p) { return ((int)p[0]<<8)+(int)p[1]; }
int read1(uchar *p) { return (int)p[0]; }

/*  set 4byte with little endian */
void setword4(unsigned char *a,unsigned int v) {
  a[0]=v&255; a[1]=(v>>8)&255; a[2]=(v>>16)&255; a[3]=(v>>24)&255;
}

/*  save an uncompressed bmp file */
int savebmp(rgb_t *pal,int x,int y,uchar *a,char *s) {
  int w,i,j,c;
  uchar header[54]={
    0x42,0x4d,0,0,0,0,0,0,0,0,54,0,0,0,40,0,
    0,0,0,0,0,0,0,0,0,0,1,0,24,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0
  };
  uchar *b;
  FILE *f=fopen(s,"wb");
  if(!f) return 0;
  w=x*3;
  if(w&3) w+=4-(w&3);
  if(!(b=malloc(w*y))) { fclose(f); error("out of memory"); }
  setword4(header+2,54+w*y);
  setword4(header+0x12,x);
  setword4(header+0x16,y);
  setword4(header+0x22,w*y);
  for(j=0;j<y;j++) for(i=0;i<x;i++) {
    c=a[i+x*j];
    b[i*3+(y-j-1)*w+2]=(pal+c)->r;
    b[i*3+(y-j-1)*w+1]=(pal+c)->g;
    b[i*3+(y-j-1)*w]=(pal+c)->b;
  }
  fwrite(header,54,1,f);
  fwrite(b,w*y,1,f);
  fclose(f);
  return 1;
}

int checkformat(uchar *p,char *s) {
  return !strncmp((char *)p+8,s,4);
}

/*  if chunk is found, return pointer to it (that is, to the start of
    the 8-byte header with id and length) */
/*  warning, this algorithm is n^2 in the number of chunks, but the number
    of chunks should always be low, so we don't care unless it becomes a
    problem */
uchar *findchunk(uchar *p,char *chunk) {
  int len=read4(p+8)+8,at=12;
  while(at<len && strncmp((char *)p+at,chunk,4)) {
    at+=read4(p+at+4)+8;
    if(at&1) at+=2-(at&1);
  }
  return at==len?0:p+at;
}

/*  extract info from bitmap header */
void bmhd(uchar *p,int *x,int *y,int *bit,int *comp) {
  uchar *pos=findchunk(p,"BMHD");
  if(!pos) error("couldn't find bmhd chunk.");
  if(read1(pos+17)) printf("warning, found masking which is not supported.\n");
  *x=read2(pos+8);
  *y=read2(pos+10);
  *bit=read1(pos+16);
  *comp=read1(pos+18);
}

void cmap(uchar *p,rgb_t *pal) {
  int i,len;
  uchar *pos=findchunk(p,"CMAP");
  if(!pos) { printf("warning, no palette found.\n"); return; }
  len=read4(pos+4);
  for(i=0;i+i+i<len;i++) (pal+i)->r=pos[8+i*3],(pal+i)->g=pos[9+i*3],(pal+i)->b=pos[10+i*3];
}

void writeto8(uchar *a,int x,int y,int to,uchar v) {
  int xn=(x+1)&(~1),xp=to%xn,yp=to/xn;
  if(xp<x && yp<y) a[xp+yp*x]=v;
}

void processbyte(uchar *p,rgb_t *pal,int x,int y,int comp,char *s) {
  uchar *pos=findchunk(p,"BODY"),*image;
  int len,at=0,to=0,j;
  if(!pos) error("couldn't find body chunk.");
  len=read4(pos+4);
  pos+=8;
  if(!(image=calloc(x*y,1))) error("out of memory");
  if(comp) {
    while(at<len) {
      j=pos[at++];
      if(j==128) continue;
      if(j>127) j-=256;
      if(j>=0) for(j++;j;j--) writeto8(image,x,y,to++,pos[at++]);
      else {
        for(j=-j+1;j;j--) writeto8(image,x,y,to++,pos[at]);
        at++;
      }
    }
  } else while(at<len) writeto8(image,x,y,to++,pos[at++]);
  savebmp(pal,x,y,image,s);
  free(image);
}

void writeto(uchar *a,int x,int y,int bp,int *to,uchar v) {
  int xn=(x+15)&(~15);
  int i,xpos=*to%xn,ypos=*to/xn/bp,bit=*to/xn%bp,at=xpos+ypos*x;
  if(at>=xn*y) return;
  for(i=0;i<8;i++) if(xpos+7-i<x && (v&(1<<i))) a[at+7-i]|=1<<bit;
  (*to)+=8;
}

void processbit(uchar *p,rgb_t *pal,int x,int y,int comp,int bit,char *s) {
  uchar *pos=findchunk(p,"BODY"),*image;
  int len,at=0,to=0,j;
  if(!pos) error("couldn't find body chunk.");
  len=read4(pos+4);
  pos+=8;
  if(!(image=calloc(x*y,1))) error("out of memory");
  if(comp) {
    while(at<len) {
      j=pos[at++];
      if(j>127) j-=256;
      if(j>=0) for(j++;j;j--) writeto(image,x,y,bit,&to,pos[at++]);
      else {
        for(j=-j+1;j;j--) writeto(image,x,y,bit,&to,pos[at]);
        at++;
      }
    }
  } else while(at<len) writeto(image,x,y,bit,&to,pos[at++]);
  savebmp(pal,x,y,image,s);
  free(image);
}

void convert(char *s,char *t) {
  FILE *f=fopen(s,"rb");
  uchar *p;
  int size,x,y,bit,comp,format;
  rgb_t pal[256];
  if(!f) error("couldn't open the file for reading.");
  fseek(f,0,SEEK_END);
  size=ftell(f);
  if(!(p=malloc(size))) { fclose(f); error("out of memory while loading file"); }
  fseek(f,0,SEEK_SET);
  fread(p,1,size,f);
  fclose(f);
  if(checkformat(p,"ILBM")) format=0;
  else if(checkformat(p,"PBM ")) format=1;
  else error("only PBM and ILBM is supported.");
  bmhd(p,&x,&y,&bit,&comp);
  cmap(p,pal);
  if(format==1) processbyte(p,pal,x,y,comp,t);
  else processbit(p,pal,x,y,comp,bit,t);
  free(p);
}

void usage() {
  puts("lbm converter v0.1 by ruben spaans in december 2011.\n");
  puts("usage: LBM infile outfile\n");
  puts("where infile is an image file in the .iff/.lbm format, and outfile is");
  puts("a 32-bit uncompressed bmp file (for now).");
}

int main(int argc,char **argv) {
  if(argc<3) usage();
  convert(argv[1],argv[2]);
  return 0;
}
